import json

import requests


def get_data(country="NL"):
    api_key = open("key.txt").read()  # API Key should be stored in src/key.txt file
    params = {'countryCode': country}
    headers = {"auth-token": api_key}
    response = requests.get("https://api.co2signal.com/v1/latest", params=params, headers=headers)
    if response.ok:
        return json.loads(response.content.decode('utf-8'))["data"]
    else:
        return "Something went wrong."


def get_current_carbon_intensity(country="NL"):
    data: dict = get_data(country=country)
    if "carbonIntensity" in data:
        return data["carbonIntensity"]
    else:
        return Exception("No carbon intensity available.")


if __name__ == '__main__':
    carbon = get_current_carbon_intensity()
    print(carbon)



import sys

import scheduler
from interface import main_menu

if __name__ == '__main__':
    print('🌱 Welcome to CAS 🌱\n')
    if not len(sys.argv) > 1:
        main_menu()
    else:
        scheduler.main()



import os
from datetime import datetime, timedelta

from PyInquirer import prompt
from crontab import CronTab
from prompt_toolkit.validation import Validator, ValidationError
from os.path import exists

from src import scheduler
from src.task import Task

user = os.getlogin()


class NumberValidator(Validator):
    def validate(self, document):
        try:
            float(document.text)
        except ValueError:
            raise ValidationError(message="Please enter a number",
                                  cursor_position=len(document.text))


class FileValidator(Validator):
    def validate(self, document):
        try:
            if not exists(document.text):
                raise ValidationError(message="File not found: " + document.text,
                                      cursor_position=len(document.text))
        except ValueError:
            raise ValidationError(message="Please enter path to file",
                                  cursor_position=len(document.text))


class DateValidator(Validator):
    def validate(self, document):
        try:
            date = datetime.strptime(document.text, '%d/%m/%y %H:%M')
            now = datetime.now()
            if date < now:
                raise ValidationError(message="Date has already passed: " + document.text,
                                      cursor_position=len(document.text))
        except ValueError:
            raise ValidationError(message="Please enter correct date and time (e.g. 28/7/23 8:00)",
                                  cursor_position=len(document.text))


welcome_question = {
    'type': 'list',
    'name': 'choice',
    'message': 'What would you like to do?',
    'choices': ["Schedule task", "View scheduled tasks", "Help", "Close"]
}
task_question = {
    'type': "input",
    "name": "task",
    "message": "Task file (path from current location)",
    "validate": FileValidator,
}
duration_question = {
    'type': "input",
    "name": "duration",
    "message": "Duration of task (in hours)",
    "validate": NumberValidator,
    "filter": lambda val: timedelta(hours=float(val))
}
deadline_question = {
    'type': "input",
    "name": "deadline",
    "message": "Deadline for task (DD/MM/YY HH:mm)",
    "validate": DateValidator,
    "filter": lambda val: datetime.strptime(val, '%d/%m/%y %H:%M')
}
start_option_question = {
    'type': "list",
    "name": "choice",
    "message": "Earliest allowed execution time",
    'choices': ["Now", "Later"]
}
start_question = {
    'type': "input",
    "name": "start",
    "message": "Earliest start time (DD/MM/YY HH:mm)",
    "validate": DateValidator,
    "filter": lambda val: datetime.strptime(val, '%d/%m/%y %H:%M')
}

help_question = {'type': 'list',
                 'name': 'choice',
                 'message': 'What\'s your question?',
                 'choices': ["What is Carbon Scheduler?", "How to process multiple files?", "Back"],
                 }


def main_menu():
    option = prompt(welcome_question).get("choice")
    if option == "Schedule task":
        schedule_task_process()
    elif option == "View scheduled tasks":
        cron = CronTab(user=user)
        scheduler.list_jobs(cron)
    elif option == "Help":
        help()
    else:
        print("Goodbye! 👋")


def schedule_task_process():
    file = prompt(task_question).get("task")
    duration = prompt(duration_question).get("duration")
    deadline = prompt(deadline_question).get("deadline")
    start = None
    if prompt(start_option_question).get("choice") == "Later":
        start = prompt(start_question).get("start")
    task = Task(duration, deadline, start)
    print("Scheduling task...")
    new_start, stats = scheduler.schedule_task(job=file, output_file="~/Desktop/output.txt", task=task)
    print("Task scheduled! 🥳\n"
          "Execution will start: ", new_start)


def help():
    option = prompt(help_question).get('choice')
    if option == "What is Carbon Scheduler?":
        print('Carbon Scheduler helps to schedule tasks in the most carbon-efficient way!')
        help()
    elif option == "How to process multiple files?":
        print("To Do")
        help()
    else:
        main_menu()



import json
import requests
from datetime import date


def get_marginal_forecast():
    """
    Get the marginal forecast carbon intensity hourly for the next day
    :return: array of predicted hourly values for the next 24 hours
    """
    api_key = open("key.txt").read()  # API Key should be stored in src/key.txt file

    response = requests.get("https://api.electricitymap.org/v3/marginal-carbon-intensity/forecast?zone='NL'",
                            headers={'Authorization': f'auth-token {api_key}'})
    print(f'Response received from the API: {response}')

    forecasted_marginal_carbon_intensity = None

    if response.status_code == 200:
        forecasted_marginal_carbon_intensity = response.json()['Response']
    else:
        # Currently, we use data in the file since we don't have the API key
        f = open('../models/dummy_data/forecasted_marginal.json')

        forecasted_marginal_carbon_intensity = json.load(f)

    results_data = [x["marginalCarbonIntensity"] for x in forecasted_marginal_carbon_intensity['forecast']]

    return results_data


def get_past_carbon_intensity_history():
    """
    Get range of past carbon intensity around last year on the same exact day as the time of execution of this program
    :return: Past carbon intensity history range
    """
    api_key = open("key.txt").read()  # API Key should be stored in src/key.txt file

    today = date.today()
    # calculate range based on today - at most 10 days range limit by API
    start_date = "2019-05-21T21:00:00Z"
    end_date = "2019-05-22T00:00:00Z"

    response = requests.get(
        f"https://api.electricitymap.org/v3/carbon-intensity/past-range?zone=NL&start={start_date}&end={end_date}",
        headers={'Authorization': f'auth-token {api_key}'})
    print(f'Response received from the API: {response}')

    past_carbon_intensity_history = None

    if response.status_code == 200:
        past_carbon_intensity_history = response.json()['Response']
    else:
        # Currently, we use data in the file since we don't have the API key
        f = open('../models/dummy_data/past_carbon_history.json')

        past_carbon_intensity_history = json.load(f)

    results_data = [x["carbonIntensity"] for x in past_carbon_intensity_history['data']]

    return results_data


print(f'Forecasted marginal intensities for the next 24 hours\n {get_marginal_forecast()}')
print(f'Last year carbon intensity around today\n {get_past_carbon_intensity_history()}')



import argparse
import datetime as dt
import os
from src.offline_model import OfflineModel
import pickle
import subprocess
import time
import croniter
from crontab import CronTab, CronItem
from src.task import Task

program_name = 'Carbon Scheduler'
pickle_file = 'carbonstats.pickle'
user = os.getlogin()


def parse_args():
    # Create the parser
    parser = argparse.ArgumentParser(description=f"{program_name} v0.1")

    # Add the arguments
    parser.add_argument('--job',
                        type=str,
                        help='the Python filename of the job to schedule')

    parser.add_argument('--repeat',
                        metavar='SCHEDULE',
                        type=str,
                        help='a cron expression or a natural language expression describing the \
                       repetition schedule of the job')

    parser.add_argument('--at',
                        metavar='DATE_SPECIFICATION',
                        type=str,
                        help='a string describing when to schedule the task')

    parser.add_argument('-t',
                        action='store_true',  # so default = False
                        help='pass this flag if you set --at and you want to specify time \
                             absolutely instead of relatively')

    parser.add_argument('--span', '--duration',
                        type=str,
                        help='the estimated span, or duration, of the job, in seconds',
                        required=False)

    parser.add_argument('--deadline',
                        type=str,
                        help='the estimated deadline of the job, in seconds from now',
                        required=False)

    parser.add_argument('--output', '--o',
                        type=str,
                        help='the path to the output file',
                        required=False)

    parser.add_argument('-green', '-g', '-mybasementisfullofbatteries', '-mbifob',
                        action='store_true',  # so default = False
                        help='pass this flag if your computing infrastructure is fully \
                        carbon-neutral')

    parser.add_argument('--list', '-l',
                        action='store_true',
                        help="list all scheduled jobs")

    parser.add_argument('--cancel_repeating', '-cr',
                        type=int,
                        help="Cancel repeating job given the id")

    parser.add_argument('--cancel_one', '-co',
                        type=int,
                        help="Cancel one-off job given the id")

    args = parser.parse_args()

    return args


def list_jobs(cron):
    print("All scheduled jobs:")
    print()
    print("Repeating jobs:")
    for (idx, job) in enumerate(cron):
        now = dt.datetime.now()
        cron = croniter.croniter(str(job.slices.render()), now)
        next_run = cron.get_next(dt.datetime)
        print(f' Id: {idx:2}| Next run: {next_run} | Job: {str(job):10}')
    print()
    cmd = 'atq'
    result = subprocess.check_output(cmd, universal_newlines=True)
    print("One-off jobs:")
    print(result)


def cancel_repeating(cron, args):
    job_removed = None
    idx = 0
    for job in cron:
        # print(f"Comparing idx {idx} with our target {args.cancel_repeating}")
        if idx == int(args.cancel_repeating):
            if isinstance(job, CronItem):
                job_removed = job
                break
        idx = idx + 1

    if job_removed is not None:
        # this doesn't actually remove the job
        cron.remove(job_removed)
        print(f'Cancelled repeating job with id: {args.cancel_repeating:2} and command: {str(job_removed):10}')


def cancel_one(args):
    subprocess.call(["at", "-r", str(args.cancel_one)], shell=False)
    print(f'Cancelled one-off job with id: {args.cancel_one:2}')


def schedule_repeating(cron, args, output_file):
    duration = dt.timedelta(seconds=3600)  # in seconds

    if args.span is not None:
        duration = dt.timedelta(seconds=int(args.span))

    deadline = dt.datetime.now() + dt.timedelta(days=7)
    if args.deadline is not None:
        deadline = dt.datetime.now() + dt.timedelta(seconds=int(args.deadline))

    model = OfflineModel('NL')
    task = Task(duration, deadline, dt.datetime.now())
    best_start, stats = model.best_24h_start_point(task)
    best_start_cron = args.repeat
    if args.repeat == "@daily":
        best_start_cron = "0 " + str(best_start.hour) + " * * *"
    elif args.repeat == "@weekly":
        best_start_cron = "0 " + str(best_start.hour) + " 1 * *"
    elif args.repeat == "@monthly":
        best_start_cron = "0 " + str(best_start.hour) + " * 1 *"
    elif args.repeat == "@yearly":
        best_start_cron = "0 " + str(best_start.hour) + " * * 1"

    # we have a repeating job, schedule using `cron`
    # extra parentheses around the strings to get a multiline string without whitespace
    # see: https://stackoverflow.com/a/10660443
    item = CronItem.from_line((f'{best_start_cron} python {os.path.abspath(args.job)} >> '
                               f'{output_file}'), cron=cron)
    cron.append(item)
    cron.write()
    next_run = croniter.croniter(args.repeat, dt.datetime.now()).get_next(dt.datetime)
    print(f'scheduled repeating job with schedule {args.repeat} - next run at {next_run}')


def schedule_task(task, job, output_file):
    model = OfflineModel('NL')
    new_start, stats = model.best_week_start_point(task)
    at_format = '%Y%m%d%H%M'  # at's format:[[CC]YY]MMDDhhmm
    time_format = '%d/%m/%Y %H:%M'
    at_options = "-t"
    at_time = new_start.strftime(at_format)
    cmd = (f'python {os.path.abspath(job)} >> {output_file} '
           f'| at {at_options} {at_time} >> /dev/null 2>&1')
    subprocess.run(cmd, shell=True)  # TODO: remove shell=True
    time = new_start.strftime(time_format)
    return time, stats


def schedule_one(args, output_file):
    duration = dt.timedelta(seconds=3600)  # in seconds

    if args.span is not None:
        duration = dt.timedelta(seconds=int(args.span))

    deadline = dt.datetime.now() + dt.timedelta(days=7)
    if args.deadline is not None:
        deadline = dt.datetime.now() + dt.timedelta(seconds=int(args.deadline))

    model = OfflineModel('NL')
    task = Task(duration, deadline, dt.datetime.now())
    new_start, stats = model.process_concurrently([task])[0]

    # we have a one-off job, schedule using `at`
    format = '%Y%m%d%H%M'  # at's format:[[CC]YY]MMDDhhmm
    if args.green is False:
        time_str = new_start.strftime(format)
    else:
        time_str = dt.datetime.now.strftime(format)
    at_options = ""
    at_time = args.at
    if args.t is True:
        at_time = time_str
        at_options = '-t'
    cmd = (f'python {os.path.abspath(args.job)} >> {output_file} '
           f'| at {at_options} {at_time} >> /dev/null 2>&1')
    subprocess.run(cmd, shell=True)  # TODO: remove shell=True

    time_string = args.at
    if args.t is True:
        time_string = new_start.strftime("%Y-%m-%d %H:%M:%S")
    print(f'🌱 Optimizing schedule to lower carbon emissions... 🌱')
    time.sleep(3)
    print(f'Scheduled one-off job for {time_string}')

    original_start = stats['original_start']
    new_start = stats['optimized_start']
    reduction = stats['reduction']
    original_cost = stats['original_cost']
    optimized_cost = stats['optimized_cost']

    totals = None
    try:
        with open(pickle_file, 'rb') as handle:
            totals = pickle.load(handle)
    except:
        data = {'original_cost': original_cost, 'optimized_cost': optimized_cost}
        try:
            with open(pickle_file, 'wb') as handle:
                pickle.dump(data, handle)
        except:
            print("Couldn't read statistics file")

        totals = data

    original_cost_total = totals['original_cost']
    optimized_cost_total = totals['optimized_cost']
    original_cost_total = original_cost_total + original_cost
    optimized_cost_total = optimized_cost_total + optimized_cost
    with open(pickle_file, 'wb') as handle:
        pickle.dump({'original_cost': original_cost_total,
                     'optimized_cost': optimized_cost_total},
                    handle)

    dateformat = "%Y-%m-%d %H:%M:%S"
    print((f'🌱 Rescheduling from {original_start.strftime(dateformat)} to {new_start.strftime(dateformat)} '
           f'will reduce CO2 intensity by {reduction:.2f}% 🌱'))
    print((f'🌱 Rescheduling has reduced your total carbon intensity from {original_cost_total}'
           f' gCO2/kWh to {optimized_cost_total} gCO2/kWh'
           f' (-{(1.0 - (float(optimized_cost_total) / float(original_cost_total))) * 100 :.2f}%) 🌱'))


def main():
    args = parse_args()

    output_file = '/tmp/cron.lst'
    if args.output is not None:
        output_file = args.output
    output_file = output_file + ' 2>&1'  # send both stdout and stderr to our output file

    cron = CronTab(user=user)

    if args.list:
        list_jobs(cron)
    elif args.cancel_repeating is not None:
        cancel_repeating(cron, args)
    elif args.cancel_one is not None:
        cancel_one(args)
    else:
        if args.repeat is not None:
            schedule_repeating(cron, args, output_file)
        elif args.at is not None:
            schedule_one(args, output_file)
        else:
            print('error: one of the following arguments is required: repeat, at')



import datetime as dt


class Task:
    def __init__(self, duration, deadline, start=None):
        self.duration = duration
        self.deadline = deadline
        self.start = start or dt.datetime.now()
        if self.deadline < self.start:
            raise Exception("Deadline has already passed." if start is None else "Deadline is before start.")
        if self.duration > (self.deadline - self.start):
            raise Exception("Not enough time to complete the task before the deadline.")



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="carbon-aware-scheduler",
    version="1.0.0",
    description="Carbon-minimizing task scheduler for Unix systems",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/abel-vs/carbon-scheduler",
    author="Wander Siemers, Abel van Steenweghen, Florentin Arsene",
    author_email="wandersiemers@me.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
    ],
    packages=["src"],
    include_package_data=True,
    install_requires=["numpy", "pandas", "requests", "python-crontab", "croniter"],
    entry_points={
        "console_scripts": [
            "cas=src.__main__:main",
        ]
    },
)


import datetime as dt
import numpy as np

from src.data import get_hour_data, get_week_data


class OfflineModel:

    def __init__(self, country, hour_data=None, week_data=None):
        self.country = country
        self.hour_data = hour_data if hour_data is not None else get_hour_data()
        self.week_data = week_data if week_data is not None else get_week_data()

    def best_24h_start_point(self, task):
        """
        Find best start point in the coming 24 hours.
        :return: Best start hour in the coming 24 hours.
        """
        costs = []
        day_cost = sum(self.hour_data)  # Cost for a whole day
        hrs = task.duration.seconds // 3600  # Costs for remaining hours
        for i in range(24):
            # Add hour costs for whole days, only for tasks longer than 24h.
            cost = day_cost * task.duration.days
            if i + hrs <= 24:
                cost += sum(self.hour_data[i:i + hrs])
            else:
                hrs_left = (i + hrs) % 24
                cost += sum(self.hour_data[i:i + (hrs - hrs_left)])
                cost += sum(self.hour_data[:hrs_left])
            costs.append(cost)

        best_start_hour = np.argmin(costs)
        best_start = dt.datetime.now().replace(hour=best_start_hour, minute=0, second=0)
        min_cost = np.min(costs)
        hour = task.start.hour
        original_cost = day_cost * task.duration.days + \
            sum(self.hour_data[hour:hour + hrs]) + \
            sum(self.hour_data[:24 - (hour + hrs)])

        stats = self.get_stats(task.start, best_start, original_cost, min_cost)

        return best_start, stats

    def best_week_start_point(self, task):
        """
        Find best start point in the coming 7 days.
        :return: Best start hour in the 7 days, .
        """

        # Structure data array to start from start hour index
        start_index = task.start.hour + task.start.weekday() * 24
        data = self.week_data.flatten()
        data = np.concatenate([data[start_index:], data[:start_index]])

        # Calculate size of search space
        time_left = task.deadline - task.start
        hours_left = time_left.total_seconds() // 3600
        # If we have more than a week left to schedule the task,
        # we find the best place in the coming week, since all weeks are the same.
        search_space = int(min(hours_left, len(data)))

        costs = []
        duration_in_hours = task.duration.seconds // 3600
        duration_in_weeks = task.duration.days // 7
        week_cost = sum(data)

        original_cost = sum(np.concatenate([data[0:], data[:0]])[:duration_in_hours])
        for i in range(search_space):
            # Sum the carbon from start point i until completion of the task
            data_from_i = np.concatenate([data[i:], data[:i]])
            cost = sum(data_from_i[:duration_in_hours])
            # If duration is longer than a whole week
            cost += week_cost * duration_in_weeks
            costs.append(cost)

        # Extract optimal start time
        min_cost = np.amin(costs)
        min_cost_hour = int(np.argmin(costs))

        # Add the number of hours for the best start time to the start date.
        best_start = (task.start + dt.timedelta(hours=min_cost_hour)).replace(minute=0, second=0)
        stats = self.get_stats(task.start, best_start, original_cost, min_cost)

        return best_start, stats

    @staticmethod
    def get_stats(original_start, optimal_start, original_cost, optimal_cost):
        reduction = (1.0 - (optimal_cost / original_cost)) * 100
        return {'reduction': reduction, 'original_cost': original_cost,
                'original_start': original_start, 'optimized_cost': optimal_cost,
                'optimized_start': optimal_start}

    def process_all_consecutive(self, tasks):
        """
        Processes all tasks consecutively in the most carbon efficient way.
        The task schedules may not overlap.
        All tasks should be handled before their deadlines.
        :param tasks: List of tasks, being tuples of a duration and a deadline.
        :return: One starting point for processing the whole batch of tasks.
        """
        pass

    def process_distributed(self, tasks):
        """
        Processes all tasks in the most carbon efficient way, without overlapping
        The task schedules don't have to be executed consecutively, meaning the program 
        may be idle during the worst hours.
        The task schedules may not overlap.
        All tasks should be handled before their deadlines.
        :param tasks: List of tasks, being tuples of a duration and a deadline.
        :return: List of starting points for each task
        """
        pass

    def process_concurrently(self, tasks):
        """
        Processes all tasks on their optimal time, possibly concurrently.
        This means for each task it searches the best possible time to execute before the deadline.
        The task schedules may overlap.
        :param tasks: List of tasks, being tuples of a duration and a deadline.
        :return: List of starting points for each task
        """
        starting_points = []
        for task in tasks:
            starting_points.append(self.best_week_start_point(task))
        return starting_points


if __name__ == '__main__':
    offline_model = OfflineModel("NL")
